@@ -1,3 +1,4 @@
+from text_explainability._version import __version__, __version_info__
 from text_explainability.data import from_string, import_data, train_test_split
 from text_explainability.global_explanation import (KMedoids,
                                                     LabelwiseKMedoids,
@@ -6,10 +7,8 @@ from text_explainability.global_explanation import (KMedoids,
                                                     TokenInformation)
 from text_explainability.local_explanation import (LIME, Anchor, KernelSHAP,
                                                    LocalRules, LocalTree)
-from text_explainability.model import from_sklearn
+from text_explainability.model import import_model
 from text_explainability.utils import (character_detokenizer,
                                        character_tokenizer,
                                        default_detokenizer, default_tokenizer,
                                        word_detokenizer, word_tokenizer)
-
-__version__ = '0.5.8'
