@@ -12,6 +12,7 @@ from typing import Callable, Optional, Sequence, Tuple, Union
 
 import numpy as np
 import six
+from genbase import Readable, SeedMixin, add_callargs
 from instancelib import (AbstractEnvironment, InstanceProvider, LabelProvider,
                          MemoryLabelProvider, TextEnvironment)
 from instancelib.instances.text import TextInstance, TextInstanceProvider
@@ -22,7 +23,6 @@ from sklearn.tree import DecisionTreeClassifier
 from .data.augmentation import LeaveOut, LocalTokenPertubator
 from .data.weights import exponential_kernel, pairwise_distances
 from .decorators import text_instance
-from .default import Readable
 from .generation.feature_selection import FeatureSelector
 from .generation.return_types import FeatureAttribution, Rules
 from .generation.surrogate import LinearSurrogate, RuleSurrogate, TreeSurrogate
@@ -51,7 +51,7 @@ def default_env(env: Optional[AbstractEnvironment] = None) -> AbstractEnvironmen
     return empty_env
 
 
-class LocalExplanation(Readable):
+class LocalExplanation(Readable, SeedMixin):
     def __init__(self,
                  env: Optional[AbstractEnvironment] = None,
                  augmenter: Optional[LocalTokenPertubator] = None,
@@ -79,7 +79,7 @@ class LocalExplanation(Readable):
                 labelset = list(self.env.labels.labelset)
         self.labelset = labelset
         self.augmenter = augmenter
-        self._seed = seed
+        self._seed = self._original_seed = seed
 
     @text_instance(tokenize=True)
     def augment_sample(self,
@@ -200,9 +200,10 @@ class LIME(LocalExplanation, WeightedExplanation):
         LocalExplanation.__init__(self, env=env, augmenter=augmenter, labelset=labelset, seed=seed)
         WeightedExplanation.__init__(self, kernel=kernel, kernel_width=kernel_width)
         if local_model is None:
-            local_model = LinearSurrogate(Ridge(alpha=1, fit_intercept=True, random_state=self._seed))
+            local_model = LinearSurrogate(Ridge(alpha=1, fit_intercept=True, random_state=self.seed))
         self.local_model = local_model
 
+    @add_callargs
     @text_instance(tokenize=True)
     def __call__(self,
                  sample: TextInstance,
@@ -212,7 +213,8 @@ class LIME(LocalExplanation, WeightedExplanation):
                  n_features: int = 10,
                  feature_selection_method: str = 'auto',
                  weigh_samples: bool = True,
-                 distance_metric: str = 'cosine') -> FeatureAttribution:
+                 distance_metric: str = 'cosine',
+                 **kwargs) -> FeatureAttribution:
         """Calculate feature attribution scores using `LIME Text`_.
 
         Args:
@@ -283,7 +285,10 @@ class LIME(LocalExplanation, WeightedExplanation):
                                   scores=feature_importances,
                                   used_features=used_features,
                                   labels=labels,
-                                  labelset=self.labelset)
+                                  labelset=self.labelset,
+                                  type='local_explanation',
+                                  method='lime',
+                                  callargs=kwargs.pop('__callargs__', None))
 
 
 class KernelSHAP(LocalExplanation):
@@ -347,12 +352,14 @@ class KernelSHAP(LocalExplanation):
             raise Exception(f'Unknown value "{l1_reg}" for l1_reg')
         return nonzero
 
+    @add_callargs
     @text_instance(tokenize=True)
     def __call__(self,
                  sample: TextInstance,
                  model: AbstractClassifier,
                  n_samples: Optional[int] = None,
-                 l1_reg: Union[int, float, str] = 'auto') -> FeatureAttribution:
+                 l1_reg: Union[int, float, str] = 'auto',
+                 **kwargs) -> FeatureAttribution:
         """Calculate feature attribution scores using `KernelShap`_.
 
         Args:
@@ -414,7 +421,10 @@ class KernelSHAP(LocalExplanation):
                                   base_score=y_null,
                                   used_features=used_features,
                                   labels=np.arange(y.shape[1]),
-                                  labelset=self.labelset)
+                                  labelset=self.labelset,
+                                  type='local_explanation',
+                                  method='kernel_shap',
+                                  callargs=kwargs.pop('__callargs__', None))
 
 
 class Anchor(LocalExplanation):
@@ -481,6 +491,7 @@ class Anchor(LocalExplanation):
 
         raise NotImplementedError('[WIP] Implementing anchor/anchor_base.py')
 
+    @add_callargs
     @text_instance
     def __call__(self,
                  sample: TextInstance,
@@ -490,7 +501,8 @@ class Anchor(LocalExplanation):
                  min_confidence: float = 0.95,
                  delta: float = 0.05,
                  epsilon: float = 0.1,
-                 max_anchor_size: Optional[int] = None):
+                 max_anchor_size: Optional[int] = None,
+                 **kwargs):
         raise NotImplementedError('Only partially implemented')
         # https://github.com/marcotcr/anchor/blob/master/anchor/anchor_text.py
         # https://github.com/marcotcr/anchor/blob/master/anchor/anchor_base.py
@@ -526,10 +538,11 @@ class LocalTree(LocalExplanation, WeightedExplanation):
         LocalExplanation.__init__(self, env=env, augmenter=augmenter, labelset=labelset, seed=seed)
         WeightedExplanation.__init__(self, kernel=kernel, kernel_width=kernel_width)
         if local_model is None:
-            local_model = TreeSurrogate(DecisionTreeClassifier(random_state=self._seed))
+            local_model = TreeSurrogate(DecisionTreeClassifier(random_state=self.seed))
         self.local_model = local_model
         self.explanation_type = explanation_type
 
+    @add_callargs
     @text_instance
     def __call__(self,
                  sample: TextInstance,
@@ -539,6 +552,8 @@ class LocalTree(LocalExplanation, WeightedExplanation):
                  distance_metric: str = 'cosine',
                  max_rule_size: Optional[int] = None,
                  **sample_kwargs):
+        callargs = sample_kwargs.pop('__callargs__', None)
+
         provider, original_id, perturbed, y = self.augment_sample(sample,
                                                                   model,
                                                                   n_samples=n_samples,
@@ -554,7 +569,10 @@ class LocalTree(LocalExplanation, WeightedExplanation):
                      original_id=original_id,
                      rules=self.local_model,
                      labelset=self.labelset,
-                     sampled=True)
+                     sampled=True,
+                     type='local_explanation',
+                     method='local_tree',
+                     callargs=callargs)
 
 
 class FactFoilMixin:
@@ -579,10 +597,11 @@ class FoilTree(FactFoilMixin, LocalExplanation, WeightedExplanation):
         LocalExplanation.__init__(self, env=env, augmenter=augmenter, labelset=labelset, seed=seed)
         WeightedExplanation.__init__(self, kernel=kernel, kernel_width=kernel_width)
         if local_model is None:
-            local_model = TreeSurrogate(DecisionTreeClassifier(random_state=self._seed))
+            local_model = TreeSurrogate(DecisionTreeClassifier(random_state=self.seed))
         self.local_model = local_model
         self.explanation_type = explanation_type
 
+    @add_callargs
     @text_instance
     def __call__(self,
                  sample: TextInstance,
@@ -593,6 +612,8 @@ class FoilTree(FactFoilMixin, LocalExplanation, WeightedExplanation):
                  distance_metric: str = 'cosine',
                  max_rule_size: Optional[int] = None,
                  **sample_kwargs):
+        callargs = sample_kwargs.pop('__callargs__', None)
+
         provider, original_id, perturbed, y = self.augment_sample(sample,
                                                                   model,
                                                                   n_samples=n_samples,
@@ -613,7 +634,10 @@ class FoilTree(FactFoilMixin, LocalExplanation, WeightedExplanation):
                      original_id=original_id,
                      rules=self.local_model,
                      labelset=labelset,
-                     sampled=True)
+                     sampled=True,
+                     type='local_explanation',
+                     method='foil_tree',
+                     callargs=callargs)
 
 
 class LocalRules(FactFoilMixin, LocalExplanation, WeightedExplanation):
@@ -631,10 +655,11 @@ class LocalRules(FactFoilMixin, LocalExplanation, WeightedExplanation):
         if local_model is None:
             local_model = RuleSurrogate(SkopeRules(max_depth_duplication=2,
                                                    n_estimators=30,
-                                                   random_state=self._seed))
+                                                   random_state=self.seed))
         self.local_model = local_model
         self.explanation_type = explanation_type
 
+    @add_callargs
     @text_instance
     def __call__(self,
                  sample: TextInstance,
@@ -644,6 +669,8 @@ class LocalRules(FactFoilMixin, LocalExplanation, WeightedExplanation):
                  weigh_samples: bool = True,
                  distance_metric: str = 'cosine',
                  **sample_kwargs):
+        callargs = sample_kwargs.pop('__callargs__', None)
+
         provider, original_id, perturbed, y = self.augment_sample(sample,
                                                                   model,
                                                                   n_samples=n_samples,
@@ -662,4 +689,7 @@ class LocalRules(FactFoilMixin, LocalExplanation, WeightedExplanation):
                      original_id=original_id,
                      rules=self.local_model,
                      labelset=labelset,
-                     sampled=True)
+                     sampled=True,
+                     type='local_explanation',
+                     method='local_rules',
+                     callargs=callargs)
